
# datasets.py
# Author: Johannes von Lindheim, Zuse-Institute Berlin, 2018
# Used for "On Intrinsic Dimension Estimation and Minimal Diffusion Maps Embeddings of Point Clouds",
# Johannes von Lindheim, master thesis (FU Berlin)
#
# This file provides functions for generating several test datasets for intrinsic dimension estimation
# and manifold learning procedures.

import numpy as np
from sklearn.preprocessing import normalize as sk_normalize
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


def make_sphere(n_pts, dim_sphere, dim_space, sigma, center=True):
    '''
    Samples uniformly from the dim_sphere-dimensional unit sphere and 
    embeds it into dim_space-dimensional space in a random direction.
    Adds Gaussian noise with sigma-parameter in the end.
    
    Parameters:
        n_pts: number of points to sample and return.
        dim_sphere: dimension of the unit-sphere-manifold to sample from.
        dim_space: dimension of the space, in which the sphere is embedded into.
        sigma: standard deviation of gaussian noise added to the points.
        center=True: whether to subtract the mean of the data from all points.
    returns: (n_pts, dim_space)-shaped two-dimensional ndarray of data points on the sphere
    '''
    # Generate n_pts zero points in dim_space-dimensional space
    data = np.zeros(shape=(n_pts, dim_space))
    
    # Generate dim_sphere+1 Gaussians and normalize, as described in http://mathworld.wolfram.com/SpherePointPicking.html    
    data[:, :dim_sphere+1] = np.random.normal(size=(n_pts, dim_sphere+1))
    data = sk_normalize(data.astype(float), axis=1, norm='l2', copy=True)
    
    # Randomly rotate data
    rot_mat = random_rotation_mat(dim_space)
    data = rot_mat.dot(data.T).T
    
    # Add gaussian noise with standard deviation sigma
    data = data + np.random.normal(scale=sigma, size=(n_pts, dim_space))
    
    # Center
    if center:
        data = data - np.mean(data, axis=0)
    
    return data
    
def make_cube(n_pts, dim_cube, dim_space, sigma, center=True):
    '''
    Samples uniformly from the dim_sphere-dimensional unit cube and 
    embeds it into dim_space-dimensional space in a random direction.
    Adds Gaussian noise with sigma-parameter in the end.
    
    Parameters:
        n_pts: number of points to sample and return.
        dim_cube: dimension of the unit-cube-manifold to sample from.
        dim_space: dimension of the space, in which the sphere is embedded into.
        sigma: standard deviation of gaussian noise added to the points.
        center=True: whether to subtract the mean of the data from all points.
    returns: (n_pts, dim_space)-shaped two-dimensional ndarray of data points on the cube
    '''
    # Draw n_pts points randomly from the dim_cube-dimensional unit cube, filled up with zeros
    data = np.random.uniform(size=(n_pts, dim_space))
    data[:, dim_cube:] = 0.0
    
    # Randomly rotate data
    rot_mat = random_rotation_mat(dim_space)
    data = rot_mat.dot(data.T).T
    
    # Add gaussian noise with standard deviation sigma
    data = data + np.random.normal(scale=sigma, size=(n_pts, dim_space))
    
    # Center
    if center:
        data = data - np.mean(data, axis=0)
    
    return data
    
def make_spherical_orthant(n_pts, dim_data, dim_space, sigma, center=True):
    '''
    Samples uniformly from the dim_sphere-dimensional unit sphere, mirors into pos. orthant
    and embeds this into dim_space-dimensional space in a random direction.
    
    Parameters:
        n_pts: number of points to sample and return.
        dim_data: intrinsic dimension of the manifold to sample from.
        dim_space: dimension of the space, in which the sphere is embedded into.
        sigma: standard deviation of gaussian noise added to the points.
        center=True: whether to subtract the mean of the data from all points.
    returns: (n_pts, dim_space)-shaped two-dimensional ndarray of data points on the sphere piece
    '''
    # Generate n_pts zero points in dim_space-dimensional space
    data = np.zeros(shape=(n_pts, dim_space))
    
    # Generate dim_sphere+1 Gaussians, mirror into positive orthant and normalize, as described in http://mathworld.wolfram.com/SpherePointPicking.html    
    data[:, :dim_data+1] = np.random.normal(size=(n_pts, dim_data+1))
    data = np.abs(data)
    data = sk_normalize(data.astype(float), axis=1, norm='l2', copy=True)
    
    # Randomly rotate data
    rot_mat = random_rotation_mat(dim_space)
    data = rot_mat.dot(data.T).T
    
    # Add gaussian noise with standard deviation sigma
    data = data + np.random.normal(scale=sigma, size=(n_pts, dim_space))
    
    # Center
    if center:
        data = data - np.mean(data, axis=0)
    
    return data
    
def make_two_spheres(n_pts, dim_spheres, dim_space, sigma, center=True):
    '''
    Samples uniformly from the dim_spheres-dimensional unit sphere and embeds two copies
    of the result into dim_space-dimensional space in a random (but equal for both copies) direction.
    Adds Gaussian noise with sigma-parameter in the end.
    
    Parameters:
        n_pts: number of points of whole dataset to sample and return. Deletes last (random) point, 
            n_pts is not even
        dim_spheres: dimension of the unit-sphere-manifold to sample from.
        dim_space: dimension of the space, in which the spheres are embedded into.
        sigma: standard deviation of gaussian noise added to the points.
        center=True: whether to subtract the mean of the data from all points.
    returns: (n_pts, dim_space)-shaped two-dimensional ndarray of data points on two spheres
    '''
    
    # Generate n_pts zero points in dim_space-dimensional space
    n_pts_one_sphere = np.ceil(n_pts/2).astype(int)
    zero_one = n_pts - 2*n_pts_one_sphere
    data = np.zeros(shape=(n_pts, dim_space))
    
    # Generate dim_spheres+1 Gaussians and normalize, as described in http://mathworld.wolfram.com/SpherePointPicking.html    
    data[:n_pts_one_sphere, :dim_spheres+1] = np.random.normal(size=(n_pts_one_sphere, dim_spheres+1))
    data = sk_normalize(data.astype(float), axis=1, norm='l2', copy=True)
    
    # Randomly rotate data
    rot_mat = random_rotation_mat(dim_space)
    data = rot_mat.dot(data.T).T
    
    # Copy sphere
    translate = max(3, np.sqrt(dim_space)) * sk_normalize(np.random.normal(size=(1, dim_space)).astype(float).reshape(1, -1), norm='l2', copy=True)
    zero_one = 2*n_pts_one_sphere - n_pts
    data[n_pts_one_sphere:, :] = data[:n_pts_one_sphere-zero_one, :] + np.tile(translate, (n_pts_one_sphere-zero_one, 1))
    
    # Add gaussian noise with standard deviation sigma
    data = data + np.random.normal(scale=sigma, size=(n_pts, dim_space))
    
    # Center
    if center:
        data = data - np.mean(data, axis=0)
    
    return data
    
def make_swiss_roll(num_pts=1000, phi_max=15, z_len=40, rand_scale=0.1):
    '''
    Calculate data points for 3D swiss roll.
    
    Parameters:
        num_pts=1000: number of points to generate
        phi_max=15: maximum angle of archimedes spiral
        z_len=1: 'length' of roll, i.e. length in direction orthogonal to angle
        rand_scale: standard deviation of noise
    returns:
        (data, angles).
        data: (num_pts, 3)-shaped ndarray with data points of swiss roll
        angles: (num_pts).shaped ndarray with the corresponding angles
    '''
    
    angles = phi_max * np.random.rand(num_pts) # random uniformly distributed angles
    noise = np.random.randn(num_pts, 3); # 3d-noise
    X = angles * np.sin(angles) + rand_scale * noise[:, 0];
    Y = angles * np.cos(angles) + rand_scale * noise[:, 1];
    Z = z_len * np.random.rand(num_pts) + rand_scale * noise[:, 2];
    
    return (np.vstack((X, Y, Z)).T, angles)

    
def make_swiss_roll_grid(n_angle_samples=50, n_z_samples=7, phi_max=15, z_len=10):
    '''
    Calculate data points for even 3D swiss roll grid.
    
    Parameters:
        n_angle_samples=200: number of grid lines in the angular direction
        n_z_samples=100: number of grid lines in z direction
        phi_max=15: maximum angle of archimedes spiral
        z_len=10: 'length' of roll, i.e. length in direction orthogonal to angle
    returns:
        (data, angles).
        data: (num_pts, 3)-shaped ndarray with data points of swiss roll
        angles: (num_pts).shaped ndarray with the corresponding angles
    '''
    
    angles = np.linspace(0, phi_max, n_angle_samples)
    zs = np.linspace(0, z_len, n_z_samples)
    points = np.array(np.meshgrid(angles, zs)).T.reshape(-1, 2)
    X = points[:, 0] * np.sin(points[:, 0])
    Y = points[:, 0] * np.cos(points[:, 0])
    Z = points[:, 1]
    
    return (np.vstack((X, Y, Z)).T, points[:, 0])

    
#############################
# Helper and test functions #
#############################
    
def test_make_sphere(n_pts, dim_sphere, sigma):
    '''
    Test make_sphere by taking n_pts points in the dim_sphere-dimensional unit sphere in 3d space using make_sphere and then plotting them.
    
    Parameters:
        n_pts: number of test points to generate
        dim_sphere: intrinsic dimension of the unit sphere to generate
        sigma: standard deviation of noise
        
    returns:
    '''
    # Generate data
    data = make_sphere(n_pts, dim_sphere, 3, sigma, center=True)
        
    # Plot
    ax = plt.subplot(projection='3d')
    plt.title('Perturbed and randomly rotated points of {0}-dimensional unit sphere'.format(dim_sphere))
    ax.scatter(data[:, 0], data[:, 1], data[:, 2], depthshade=True)
    ax.view_init(60, 15)
    ax.set_aspect('equal')
    
def test_make_cube(n_pts, dim_cube, sigma):
    '''
    Test make_cube by taking n_pts points in the dim_sphere-dimensional unit cube in 3d space using make_sphere and then plotting them.
    
    Parameters:
        n_pts: number of test points to generate
        dim_cube: intrinsic dimension of the unit sphere to generate
        sigma: standard deviation of noise
        
    returns:
    '''
    # Generate data
    data = make_cube(n_pts, dim_cube, 3, sigma, center=True)
        
    # Plot
    ax = plt.subplot(projection='3d')
    plt.title('Perturbed and randomly rotated points of {0}-dimensional unit cube'.format(dim_cube))
    ax.scatter(data[:, 0], data[:, 1], data[:, 2], depthshade=True)
    ax.view_init(60, 15)
    ax.set_aspect('equal')
    
def test_make_spherical_orthant(n_pts, dim_data, sigma):
    '''
    Test make_spherical_orthant by taking n_pts points by sampling in 3d space using make_spherical_orthant and then plotting them.
    
    Parameters:
        n_pts: number of test points to generate
        dim_data: intrinsic dimension of the manifold to sample from
        sigma: standard deviation of noise
        
    returns:
    '''
    # Generate data
    data = make_spherical_orthant(n_pts, dim_data, 3, sigma, center=True)
        
    # Plot
    ax = plt.subplot(projection='3d')
    plt.title('Perturbed and randomly rotated points of {0}-dimensional unit sphere'.format(dim_data))
    ax.scatter(data[:, 0], data[:, 1], data[:, 2], depthshade=True)
    ax.view_init(60, 15)
    ax.set_aspect('equal')
    
def test_make_two_spheres(n_pts, dim_spheres, sigma):
    '''
    Test make_two_spheres by taking n_pts points in the dim_spheres-dimensional unit sphere in 3d space using make_two_spheres and then plotting them.
    
    Parameters:
        n_pts: number of test points to generate
        dim_spheres: intrinsic dimension of the unit sphere to generate
        sigma: standard deviation of noise
        
    returns:
    '''
    # Generate data
    data = make_two_spheres(n_pts, dim_spheres, 3, sigma, center=True)
        
    # Plot
    ax = plt.subplot(projection='3d')
    plt.title('Two perturbed and randomly rotated points of {0}-dimensional unit spheres'.format(dim_spheres))
    ax.scatter(data[:, 0], data[:, 1], data[:, 2], depthshade=True)
    ax.view_init(60, 15)
    ax.set_aspect('equal')
    
def test_make_swiss_roll_grid(n_angle_samples=50, n_z_samples=7, phi_max=15, z_len=10):
    '''
    Test make_swiss_roll_grid by plotting one swiss roll.
    
    Parameters:
        n_pts: number of test points to generate
        dim: intrinsic dimension of the unit sphere to generate
        sigma: standard deviation of noise
        
    returns:
    '''
    # Generate data
    data, angles = make_swiss_roll_grid(n_angle_samples=n_angle_samples, n_z_samples=n_z_samples, phi_max=phi_max, z_len=z_len)
        
    # Plot
    ax = plt.subplot(projection='3d')
    plt.title('Swiss roll')
    ax.scatter(data[:, 0], data[:, 1], data[:, 2], c=angles, cmap=plt.cm.Spectral, depthshade=True)
    ax.view_init(78, 15)
    ax.set_aspect('equal')

def random_rotation_mat(d):
    '''
    Generate a dxd uniformly random rotation matrix in d-dimensional space (only rotates in upper half!).
    
    Parameters:
        d: dimension of rotation matrix
        
    returns: (d, d)-shaped rotation matrix (ndarray)
    '''
    # Generate dxd-matrix with each entry sampled from a normal distribution
    mat = np.random.normal(size=(d, d))
    
    # Do QR decomposition
    Q, R = np.linalg.qr(mat)
    
    # Q is a uniformly random rotation matrix (almost)
    return Q

def test_random_rotation_mat(n_pts):
    '''
    Test random_rotation_mat by taking n_pts times the unit vector e_1 or -e_1 in 3d space, rotating and then plotting them.
    
    Parameters:
        d: dimension of rotation matrix
        n_pts: number of test points to randomly rotate
        
    returns:
    '''
    # Generate data
    data = np.zeros((n_pts, 3))
    for i in range(n_pts):
        vec = np.zeros(3)
        vec[0] = 2*np.random.randint(0, 2) - 1 # take e_1 or -e_1 with prob. 1/2 each
        mat = random_rotation_mat(3)
        vec = mat.dot(vec)
        data[i, :] = vec
        
    # Plot
    ax = plt.subplot(projection='3d')
    plt.title('Randomly rotated points')
    ax.scatter(data[:, 0], data[:, 1], data[:, 2], depthshade=True)
    ax.view_init(60, 15)
    ax.set_aspect('equal')
