
# manifold_learn.py
# Author: Johannes von Lindheim, Zuse-Institute Berlin, 2018
# Used for "On Intrinsic Dimension Estimation and Minimal Diffusion Maps Embeddings of Point Clouds",
# Johannes von Lindheim, master thesis (FU Berlin)
#
# This file provides some functionality for performing the minimal diffusion maps algorithm, as well as the standard
# diffusion maps algorithm, and some plotting functionality.

from __future__ import absolute_import, division, print_function, unicode_literals
__metaclass__ = type

import numpy as np
import scipy.sparse as sps
from scipy.spatial import distance
from scipy.stats.mstats import gmean
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import normalize
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d import proj3d
from matplotlib.patches import FancyArrowPatch
import networkx as nx
from copy import deepcopy
import time
import sys


class ManifoldLearner(object):

    def __init__(self, data, eps, r, alpha=1.0, fit_eps_nn=True, compute_kernel=True, verbose=True):

        '''
        Manifold Learning class. Provides standard diffusion maps algorithm, as well as own adaption to solve higher harmonics problems.

        Parameters:
            data: (n_pts, d)-shaped numpy-ndarra, data to perform manifold learning on.
            eps: epsilon to use in RBF kernel, which also has influence on eps-nearest-neighbors.
            r: radius for eps-nearest-neighbors.
            alpha=1.0: How much to divide out the sample densities of data points.
            fit_eps_nn=True: If True is given, the eps-NN-object is fitted directly at initialization of the object.
        '''

        self._data = data
        self._eps = eps
        self._r = r
        self._alpha = alpha
        self._neigh_obj = None
        self._kernel = None
        self._n_pts = data.shape[0]
        self._extr_dim = data.shape[1]
        self._embedding = None
        self._eigenvals = None
        self._residuals = None
        self._verbose = verbose

        if fit_eps_nn:
            self._fit_eps_nn()
        if compute_kernel:
            self._compute_kernel()

    @property
    def data(self):
        return self._data
    @property
    def eps(self):
        return self._eps
    @property
    def r(self):
        return self._r
    @property
    def alpha(self):
        return self._alpha
    @property
    def neigh_obj(self):
        return self._neigh_obj
    @property
    def kernel(self):
        return self._kernel
    @property
    def n_pts(self):
        return self._n_pts
    @property
    def extr_dim(self):
        return self._extr_dim
    @property
    def embedding(self):
        return self._embedding
    @property
    def eigenvals(self):
        return self._eigenvals
    @property
    def residuals(self):
        return self._residuals
    @property
    def verbose(self):
        return self._verbose

    ###################
    # Main algorithms #
    ###################

    def fit_diffusion_maps(self, k, h, symmetrize=True, return_histories=False):
        '''
        Computes diffusion maps embedding with k coordinates (can be recieved from self.embedding),
        where directions of already computed coordinates are projected out.
        Caution: Eps-NN-object is only fitted new, if it does not exist yet.

        Parameters:
            k: number of diffusion coordinates to compute (must not be greater than number of data points)
            h: the step size for calculating the gradients of diffusion coordinates
                with second order finite differences
            symmetrize: whether to symmetrize the local (=> global) distances
            return_histories=True: If set to True, the local dists, global dists and difference vectors
                will be saved in all stages and returned

        returns:
            [if symmetrize == True] (local_dists_history, global_dists_history, diff_vec_history, pos):
            Local distance array, global distance arrays and difference vectors will be saved in all stages
            stages of the algorithm and returned as lists of these arrays. pos necessary for reading diff_vec_history
        '''

        if self.verbose:
            print("Initialize")

        # epsilon-nearest-neighbors
        if self._neigh_obj is None:
            self._fit_eps_nn()

        # Initialize distances, arrays, etc.
        local_dists = self.neigh_obj.radius_neighbors_graph(mode='distance')
        old_local_dists = deepcopy(local_dists)
        connectivity = self.neigh_obj.radius_neighbors_graph(mode='connectivity')
        dijkstra_cutoff = self._cutoff()
        self._residuals = np.zeros(k)
        self._residuals[0] = local_dists.sum()
        pos = connectivity.indptr
        row = np.sort(local_dists.tocoo().row)
        col = local_dists.tocoo().col
        for i in range(self.n_pts):
            col[pos[i]:pos[i+1]] = np.sort(col[pos[i]:pos[i+1]]) # sort col-vector
        zero_dists = sps.dok_matrix((self.n_pts, self.n_pts)) # array, which remembers, which distances are exactly zero
        # (needed because of sparse matrices do not necessarily save zeros)
        global_dists = deepcopy(local_dists)
        last_eig = np.zeros((self.n_pts, 1))
        last_ev = 0.0
        gradients = np.zeros((self.n_pts, self.extr_dim))
        diff_vec_arr = self._diff_vecs(connectivity)
        self._embedding = np.zeros((self.n_pts, k+1))
        self._eigenvals = np.zeros(k+1)
        # ...For every eigenvector (k), there is a extr_dim-dimensional gradient vector for
        # every point. I just do not need last gradient, so there are just k - 1 grad fields
        self._embedding[:, 0] = 1.0 # first eigenvector is the constant one...
        self._eigenvals[0] = 1.0 # ...with eigenvalue 1

        if return_histories:
            local_dist_history = [deepcopy(local_dists)]
            global_dist_history = [deepcopy(global_dists)]
            diff_vec_history = [deepcopy(diff_vec_arr)]

        # For all components to compute...
        for k in range(1, k + 1):

            if self.verbose:
                print("\rk = {0}".format(k))

            # Subtract distances in directions of first (computed) eigenvectors, if some have been computed already
            if k > 1:

                if self.verbose:
                    print("Update local distances")

                gradients = self.gradients(last_eig, last_ev, h, normalized=True)

                for i in range(self.n_pts):
                    diff_vec_arr[pos[i]:pos[i+1], :] -= np.dot(np.dot(diff_vec_arr[pos[i]:pos[i+1], :], gradients[i, :].reshape(self.extr_dim, 1)), gradients[i, :].reshape(1, self.extr_dim))
                local_dists = sps.csr_matrix((np.linalg.norm(diff_vec_arr, axis=1), (row, col)), shape=(self.n_pts, self.n_pts))
                self.residuals[k-1] = local_dists.sum()
                zero_dists = sps.csr_matrix((np.isclose(np.linalg.norm(diff_vec_arr, axis=1), 0.0), (row, col)), shape=(self.n_pts, self.n_pts)) + sps.identity(self.n_pts)

                # Symmetrize, if requested
                if symmetrize:
                    local_dists = (local_dists + local_dists.T) / 2.0
                    zero_dists = zero_dists.astype(float) + zero_dists.T
                    zero_dists[zero_dists == 1.0] = 0.0 # should also work, when symmetrize == False...
                    zero_dists = zero_dists / 2.0 # scale, s.t. every positive entry is 1.0. Needed later when adding this

                # We do not want to not add multiple edges later, when adding edges at zero_dists-1-entries
                local_dists = local_dists.tolil()
                local_dists[zero_dists == 1.0] = 0.0
                local_dists = local_dists.tocsr()

                if self.verbose:
                    print("Compute global distances")

                # Compute new global distances by applying Dijkstra with cutoff = exp(-r) (which is the cutoff
                # of diffusion maps) => path local distances together

                # First, decrease the cutoff a bit (by the average decrease factor of the distances)
                mask = (local_dists > 0.0)
                dijkstra_cutoff *= gmean(sps.csr_matrix(local_dists[mask]).data / sps.csr_matrix(old_local_dists[mask]).data)
                old_local_dists = deepcopy(local_dists)

                # Construct graph
                G = nx.from_scipy_sparse_matrix(local_dists)
                zero_dists = zero_dists.tocoo()
                G.add_weighted_edges_from(zip(zero_dists.row, zero_dists.col, [0.0]*zero_dists.nnz)) # add remembered zero edges
                start = time.clock()

                # Dijkstra
                lengths = nx.all_pairs_dijkstra_path_length(G, cutoff=dijkstra_cutoff)
                if self.this_is_authors_laptop():
                        length_dict = {node:{to_node:{'weight':dist} for (to_node, dist) in dists.items()} for (node, dists) in lengths.items()}
                else:
                    length_dict = {node:{to_node:{'weight':dist} for (to_node, dist) in dists.items()} for (node, dists) in lengths}
                end = time.clock()
                if self.verbose:
                    print("Time for global Dijkstra: {:.4f}".format(end - start))
                global_dists = nx.to_scipy_sparse_matrix(nx.Graph(length_dict))

                # Possibly save distances and difference vectors of this step
                if return_histories:
                    local_dist_history.append(deepcopy(local_dists))
                    global_dist_history.append(deepcopy(global_dists))
                    diff_vec_history.append(deepcopy(diff_vec_arr))

            if self.verbose:
                print("Compute diffusion coordinate from distances")

            # Diffusion
            kernel_data = np.exp(-(global_dists.data**2)/self.eps)
            kernel = sps.csr_matrix((kernel_data, global_dists.indices, global_dists.indptr), shape=(self.n_pts, self.n_pts))
            #kernel = kernel + zero_dists # add ones to places, where there is zero distance. Identity is automatically added
            # No need for this anymore, because networkx automatically puts 0-entries into the sparse matrix, where there
            # is a 0-path, and no entries in the matrix, where there is no path

            # Divide out density
            densities = sps.csr_matrix.sum(kernel, axis=0)
            density_mat = sps.spdiags(np.power(densities, -self.alpha), 0, self.n_pts, self.n_pts)
            kernel = density_mat.dot(kernel.dot(density_mat))

            # Row normalize
            normalize(kernel, axis=1, norm='l1', copy=False)

            # Compute next eigenvector
            eigenvals, eigenvecs = sps.linalg.eigs(kernel, k=1+1, which='LM') # account for constant vector
            last_eig[:, 0] = self.n_pts * normalize(eigenvecs.real, axis=0, norm='l1', copy=True)[:, 1]
            last_ev = eigenvals[1].real
            self.eigenvals[k] = last_ev
            self.embedding[:, k] = last_eig[:, 0]

        if return_histories:
            return local_dist_history, global_dist_history, diff_vec_history, pos


    def fit_standard_diffusion_maps(self, k):
        '''
        Computes standard diffusion maps embedding with k coordinates (can be recieved from self.embedding)
        Caution: Eps-NN-object and diffusion kernel are only fitted new, if it does not exist yet.

        Parameters:
            k: number of diffusion coordinates to compute (must not be greater than number of data points)
        '''

        # epsilon-nearest-neighbors
        if self._neigh_obj is None:
            self._fit_eps_nn()
        if self._kernel is None:
            self._compute_kernel()

        # Divide out density
        densities = sps.csr_matrix.sum(self._kernel, axis=0)
        density_mat = sps.spdiags(np.power(densities, -self._alpha), 0, self._n_pts, self._n_pts)
        self._kernel = density_mat.dot(self._kernel.dot(density_mat))

        # Row normalize
        normalize(self._kernel, axis=1, norm='l1', copy=False)

        # Compute embedding
        self._eigenvals, self._embedding = sps.linalg.eigs(self._kernel, k=k+1, which='LM') # account for constant vector

    def gradients(self, coordinate, eigenvalue, h, normalized=False):
        '''
        Computes gradients of some given (diffusion maps) embedding coordinate at all data points by extending the
        embedding coordinate (eigenvector) with the Nyström extension and then using second order finite differences.

        Parameters:
            coordinate: the (diffusion maps) embedding coordinate (eigenvector) to compute the gradients for.
            eigenvalue: the corresponding eigenvalue to the coordinate
            h: (isotropic) step size for calculating the second order finite differences
            normalized: whether to normalize the gradients to unit length. Useful, if we are only interested in the
                direction of the gradient and not the steepness. Moreover, gradients can tend to be very short.

        returns:
            (n_pts, extr_dim)-shaped ndarray, gradients
        '''

        f_minus_plus_h = np.zeros((2, self.n_pts, self.extr_dim))
        H = 1/h

        for i in range(2):
            for d in range(self.extr_dim):
                f_minus_plus_h[i, :, d] = np.squeeze(self.nystroem_extension(coordinate, np.array([eigenvalue]), self.data - h*np.power(-1, i)*np.tile(self._ek(d, self.extr_dim), (self.n_pts, 1))).real)
        result = (f_minus_plus_h[1, :, :] - f_minus_plus_h[0, :, :]) * H / 2 # central difference
        return normalize(result, axis=1, norm='l2', copy=True) if normalized else result

    def nystroem_extension(self, embedding, eigenvals, points_ext):
        '''
        Computes the Nyström extension of the computed diffusion coordinate k to the given points ext_points.

        Parameters:
            embedding: (n_pts, k)-shaped ndarray, computed diffusion coordinates to extend.
            eigenvals: corresponding eigenvalues.
            points_ext: the positions of the points, for which to compute the extension.

        returns:
            (n_ext_points, k)-shaped ndarray, Nyström extension of given points and k given embedding coordinates
        '''

        assert self._neigh_obj is not None, "self._neigh_obj is None, but neighbor-object has to be computed already"
        assert self._kernel is not None, "self._kernel is None, but kernel has to be computed already"

        n_pts_ext = points_ext.shape[0]
        if len(embedding.shape) == 1:
            embedding = embedding.reshape((-1, 1))
        k = embedding.shape[1]

        # Create diffusion of extension points to data points
        graph_ext = self._neigh_obj.radius_neighbors_graph(points_ext, mode='distance')
        kernel_data_ext = np.exp(-(graph_ext.data**2)/self._eps)
        kernel_ext = sps.csr_matrix((kernel_data_ext, graph_ext.indices, graph_ext.indptr),
                                shape=(n_pts_ext, self._n_pts))

        # Divide out densities of data points from the extension point values
        data_densities = sps.csr_matrix.sum(self._kernel, axis=0)
        data_density_mat = sps.spdiags(np.power(data_densities, -self._alpha), 0, self._n_pts, self._n_pts)
        ext_densities = sps.csr_matrix.sum(kernel_ext, axis=1)
        ext_density_mat = sps.spdiags(np.power(np.squeeze(ext_densities), -self._alpha), 0, n_pts_ext, n_pts_ext)
        kernel_ext = ext_density_mat.dot(kernel_ext.dot(data_density_mat))
        
        # Row normalize
        normalize(kernel_ext, axis=1, norm='l1', copy=False)

        # Compute extended eigenvectors
        return kernel_ext.toarray().dot(embedding).dot(np.diagflat(1/eigenvals)) # IS THIS THE RIGHT ORDER OF MATRICES?


    ####################
    # Helper functions #
    ####################

    def this_is_authors_laptop(self):
        return all((sys.version_info.major == 3, sys.version_info.minor == 5, sys.version_info.micro == 1))

    def _fit_eps_nn(self, cutoff=None):
        '''
        Fits the own epsilon-nearest-neighbor-object onto the data.
        '''
        if cutoff is None:
            cutoff = self._cutoff()

        self._neigh_obj = NearestNeighbors(radius=cutoff, algorithm='auto')
        self._neigh_obj.fit(self._data)

    def _compute_kernel(self):
        '''
        Computes the diffusion kernel from the already fitted eps-NN-object.
        '''
        assert self._neigh_obj is not None, "self._neigh_obj is None, but neighbor-object has to be computed already"

        graph = self._neigh_obj.radius_neighbors_graph(mode='distance')

        # Diffusion
        kernel_data = np.exp(-(graph.data**2)/self._eps)
        self._kernel = sps.csr_matrix((kernel_data, graph.indices, graph.indptr),
                                shape=(self._n_pts, self._n_pts)) + sps.identity(self._n_pts)


    def _ek(self, k, d):
        '''
        Unit k'th d-dimensional unit vector (Caution: indices start with 0)
        '''
        vec = np.zeros(d)
        vec[k] = 1
        return vec

    def _cutoff(self):
        '''
        Computes eps-NN cutoff from current eps and r
        '''
        return np.sqrt(self.eps*self.r)

    def _diff_vecs(self, conn):
        '''
        Computes concatenated array of difference vectors of the data vectors. The result contains exactly the difference
        vectors, where there is a 1-entry in the given (sparse) connectivity matrix.
        '''
        result = np.zeros((conn.nnz, self.extr_dim))
        counter = 0
        for i in range(self.n_pts):
            n_neighbors = conn[i, :].sum().astype(int)
            result[counter:counter+n_neighbors] = np.squeeze(self.data[np.squeeze(conn[i, :].astype(bool).toarray()), :]
                                                             - self.data[i, np.newaxis, :])
            counter += n_neighbors
        return result


    ######################
    # Plotting functions #
    ######################

    def save_if_path_given(self, save_path):
        if save_path is not None:
            plt.savefig(save_path + ".png", transparent = True)#, bbox_inches=extent)
            plt.savefig(save_path + ".pdf", transparent = True)#, bbox_inches=extent)

    def plot_residuals(self, save_path=None):
        '''
        Plot the residuals (sum of remaining local distances) for the
        '''
        plt.plot(np.arange(len(self.residuals)), 100*self.residuals/self.residuals[0], marker='o')
        ax = plt.gca()
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.set_ylim([0, 100])
        plt.xlabel("Step")
        plt.ylabel("Residual in %")
        self.save_if_path_given(save_path)
        plt.show()

    ### 2D ###

    def plot_embedding_2d(self, k1, k2, func_vals, alpha=1.0, save_path=None):
        '''
        Plot the coordinate k1 against the coordinate k2 (scatter plot).

        Parameters:
            k1: coordinate plotted on the x-axis.
            k2: coordinate plotted on the y-axis.
            func_vals: determines the color of the data points. Has to be a vector of the same length as embedding.
                If None is given, all points will be plotted in blue.
            alpha=1.0: alpha-value of data points.

        returns:
        '''

        if func_vals is not None:
            assert self._embedding.shape[0] == func_vals.shape[0], "func_vals has to be a vector of the same length as embedding."

        if func_vals is not None:
            plt.scatter(self._embedding[:, k1].real, self._embedding[:, k2].real, c=func_vals, alpha=alpha, cmap=plt.cm.Spectral)
        else:
            plt.scatter(self._embedding[:, k1].real, self._embedding[:, k2].real, alpha=alpha)
        #plt.title('Diffusion coordinate $\\psi_{0}$ plotted against $\\psi_{1}$'.format(k1, k2))
        plt.axis('tight')
        plt.xlabel('$\\psi_{0}$'.format(k1),fontsize=16)
        plt.ylabel('$\\psi_{0}$'.format(k2),fontsize=16)
        self.save_if_path_given(save_path)
        plt.show()

    def plot_coordinates_2d(self, k, save_path=None):
        '''
        Plot the first k diffusion coordinates (except for zero^th, constant one) by plotting the data points in 2d space and using a colormap to visualize diffusion coordinate.

        Parameters:
            k: how many coordinates to visualize.

        returns:
        '''

        assert self._n_pts == self._embedding.shape[0], "data and embedding vectors have to be of same length"
        assert self._embedding.shape[1] > k, "k is larger than number of given embedding vectors to visualize"

        #suptitle = 'First {0} diffusion coordinates'.format(k)
        subtitles = ['$\\psi_{0}$'.format(i+1) for i in range(k)]
        suptitle = ""

        self.plot_functions_2d(self._data, self._embedding[:, 1:], suptitle, subtitles)
        self.save_if_path_given(save_path)
        plt.show()

    def plot_functions_2d(self, data, f, suptitle, subtitles):
        '''
        Plots k 1d-functions over the data points in 2d space using a colormap.

        Parameters:
            data: data points to plot in 2d
            f: (n_pts, k)-shaped array of the function values to visualize.
            suptitle: title of whole plot
            subtitles: titles for the subplots

        returns:
        '''

        # Init plotting arrangement
        n_pts = data.shape[0]
        k = f.shape[1]
        num_subplts_per_row = 3
        num_rows = np.ceil(k / num_subplts_per_row).astype(int)
        num_cols = min(num_subplts_per_row, k)
        fig = plt.figure(figsize=(17, 6*num_rows))
        fig.suptitle(suptitle)

        # Plot
        for i in range(k):
            cur_x = i % num_subplts_per_row
            cur_y = i // num_subplts_per_row

            if i < k:
                ax = fig.add_subplot(num_rows, num_cols, i+1)
                ax.scatter(data[:, 0], data[:, 1], c=f[:, i].real, cmap=plt.cm.Spectral)
                ax.set_aspect('equal')
                ax.set_title(subtitles[i])

    def plot_single_coordinate_2d(self, k, save_path=None):
        '''
        Plot diffusion coordinate by plotting the data points in 2d space and using a colormap to visualize diffusion coordinate.

        Parameters:
            k: which coordinate to visualize.

        returns:
        '''

        ax = plt.subplot()
        plt.title('$\\psi_{0}$'.format(k))
        ax.scatter(self._data[:, 0], self._data[:, 1], c=self._embedding[:, k].real, cmap=plt.cm.Spectral)
        ax.set_aspect('equal')

        self.save_if_path_given(save_path)
        plt.show()

    ### 3D ###

    # Arrow class
    class Arrow3D(FancyArrowPatch):
        def __init__(self, xs, ys, zs, *args, **kwargs):
            FancyArrowPatch.__init__(self, (0,0), (0,0), *args, **kwargs)
            self._verts3d = xs, ys, zs

        def draw(self, renderer):
            xs3d, ys3d, zs3d = self._verts3d
            xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
            self.set_positions((xs[0],ys[0]),(xs[1],ys[1]))
            FancyArrowPatch.draw(self, renderer)

    def plot_embedding_3d(self, k1, k2, k3, func_vals, view_a=78, view_b=15, alpha=1.0, save_path=None):
        '''
        Plot the coordinates k1, k2 and k3 against each other in a 3D scatter plot.

        Parameters:
            k1: coordinate plotted on the x-axis.
            k2: coordinate plotted on the y-axis.
            k3: coordinate plotted on the z-axis.
            func_vals: determines the color of the data points. Has to be a vector of the same length as embedding.
                If None is given, all points will be plotted in blue.
            view_a: longitude in degrees.
            view_b: latitide in degrees.
            alpha=1.0: alpha-value of data points.

        returns:
        '''

        if func_vals is not None:
            assert self._embedding.shape[0] == func_vals.shape[0], "func_vals has to be a vector of the same length as embedding."


        ax = plt.subplot(projection='3d')

        if func_vals is not None:
            ax.scatter(self._embedding[:, k1].real, self._embedding[:, k2].real, self._embedding[:, k3].real, c=func_vals, alpha=alpha, cmap=plt.cm.Spectral)
        else:
            ax.scatter(self._embedding[:, k1].real, self._embedding[:, k2].real, self._embedding[:, k3].real, alpha=alpha)

        ax.view_init(view_a, view_b)
        ax.set_aspect('equal')
        #plt.xlabel('$\\psi_{0}$'.format(k1),fontsize=16)
        #plt.ylabel('$\\psi_{0}$'.format(k2),fontsize=16)
        #plt.zlabel('$\\psi_{0}$'.format(k3),fontsize=16)
        self.save_if_path_given(save_path)
        plt.show()

    def plot_coordinates_3d(self, k, view_a=78, view_b=15, save_path=None):
        '''
        Plot the first k diffusion coordinates (except for zero^th, constant one) by plotting the data points in 3d space and using a colormap to visualize diffusion coordinate.

        Parameters:
            k: how many coordinates to visualize.
            view_a: longitude in degrees.
            view_b: latitide in degrees.
            save_path: path to save figure to. If None, nothing is saved.

        returns:
        '''

        assert self._n_pts == self._embedding.shape[0], "data and embedding vectors have to be of same length"
        assert self._embedding.shape[1] > k, "k is larger than number of given embedding vectors to visualize"

        #suptitle = 'First {0} diffusion coordinates'.format(k)
        subtitles = ['$\\psi_{0}$'.format(i+1) for i in range(k)]
        suptitle = ""

        self.plot_functions_3d(self._data, self._embedding[:, 1:k+1], suptitle, subtitles, view_a, view_b)
        self.save_if_path_given(save_path)
        plt.show()

    def plot_functions_3d(self, data, f, suptitle, subtitles, view_a=78, view_b=15):
        '''
        Plots k 1d-functions over the data points in 3d space using a colormap.

        Parameters:
            data: data points to plot in 3d
            f: (n_pts, k)-shaped array of the function values to visualize.
            suptitle: title of whole plot
            subtitles: titles for the subplots
            view_a: longitude in degrees.
            view_b: latitide in degrees.
            save_path: path to save figure to. If None, nothing is saved.

        returns:
        '''

        # Init plotting arrangement
        n_pts = data.shape[0]
        k = f.shape[1]
        num_subplts_per_row = 3
        num_rows = np.ceil(k / num_subplts_per_row).astype(int)
        num_cols = min(num_subplts_per_row, k)
        fig = plt.figure(figsize=(17, 6*num_rows))
        fig.suptitle(suptitle)

        # Plot
        for i in range(k):
            ax = fig.add_subplot(num_rows, num_cols, i+1, projection='3d')
            ax.scatter(data[:, 0], data[:, 1], data[:, 2], c=f[:, i].real, cmap=plt.cm.Spectral, depthshade=True)
            ax.view_init(view_a, view_b)
            ax.set_aspect('equal')
            ax.set_title(subtitles[i])

    def plot_single_coordinate_3d(self, k, save_path=None):
        '''
        Plot diffusion coordinate by plotting the data points in 3d space and using a colormap to visualize diffusion coordinate.

        Parameters:
            k: which coordinate to visualize.

        returns:
        '''

        ax = plt.subplot(projection='3d')
        #plt.title('$\\psi_{0}$'.format(k))
        ax.scatter(self._data[:, 0], self._data[:, 1], self._data[:, 2], c=self._embedding[:, k].real, cmap=plt.cm.Spectral, depthshade=True)
        ax.view_init(78, 15)
        ax.set_aspect('equal')

        self.save_if_path_given(save_path)
        plt.show()

    def plot_3d_gradients(self, grad_arr, how_many, coordinate, lat=30, long=80, scale=5, save_path=None):
        '''
        Visualize a given gradients array of a (diffusion maps) coordinate in a 3d scatter plot using a colormap
        and arrows for the gradients of some randomly selected points.
        Most is copied from StackOverflow:
        https://stackoverflow.com/questions/22867620/putting-arrowheads-on-vectors-in-matplotlibs-3d-plot

        Parameters:
            grad_arr: array of gradients to visualize
            how_many: how many arrows to plot (needs to be less or equal to the number of points)
            coordinate: the function
        returns:
        '''

        assert how_many <= self.n_pts, "how_many needs to be smaller or equal to the number of data points"

        # Scale appropriately
        max_ax_diam = max(max(self.data[:, 0])-min(self.data[:, 0]), max(self.data[:, 1])-min(self.data[:, 1]), max(self.data[:, 2])-min(self.data[:, 2]))
        grad_scale = max_ax_diam/scale

        # Choose points, which get an arrow
        pt_indices = np.random.choice(np.arange(self.n_pts), size=how_many, replace=False)

        # Plot
        fig = plt.figure(figsize=(10, 10))
        ax = fig.add_subplot(111, projection='3d')
        ax.scatter(self.data[:, 0], self.data[:, 1], self.data[:, 2], c=coordinate, cmap=plt.cm.Spectral, depthshade=True)
        for ind in pt_indices:
            arrow = Arrow3D([self.data[ind, 0], self.data[ind, 0] + grad_scale*grad_arr[ind, 0]],
                            [self.data[ind, 1], self.data[ind, 1] + grad_scale*grad_arr[ind, 1]],
                            [self.data[ind, 2], self.data[ind, 2] + grad_scale*grad_arr[ind, 2]],
                            mutation_scale=20, lw=3, arrowstyle="-|>", color="r")
            ax.add_artist(arrow)
        ax.view_init(lat, long)
        ax.set_aspect('equal')

        self.save_if_path_given(save_path)
        plt.show()


    def plot_neighborhoods(self, point_ind, n_show, n_subplots_per_row, global_dist_history, diff_vec_history, pos, view_a, view_b, plot_new_positions=False, plot_local_dists=False, save_path=None):

        # Init plotting arrangement
        k = n_show
        num_subplts_per_row = n_subplots_per_row
        num_rows = np.ceil(k / num_subplts_per_row).astype(int)
        num_cols = min(num_subplts_per_row, k)
        fig = plt.figure(figsize=(17, 6*num_rows))

        # Scale arrow appropriately
        scale = 0.2
        max_ax_diam = max(max(self.data[:, 0])-min(self.data[:, 0]), max(self.data[:, 1])-min(self.data[:, 1]), max(self.data[:, 2])-min(self.data[:, 2]))
        grad_scale = max_ax_diam*scale

        #suptitle = 'Neighbors of data point {0} of first {1} diffusion coordinates'.format(point_ind, k)
        suptitle = ""
        subtitles = ['$\\psi_{0}$'.format(i+1) for i in range(k)]
        fig.suptitle(suptitle)

        neighbors_inds = None
        neighbors_inds_list = None
        if plot_local_dists:
            neighbors_inds = np.sort(global_dist_history[0][point_ind, :].tocoo().col)
        else:
            neighbors_inds_list = [None] * k
            for i in range(k):
                neighbors_inds_list[i] = np.sort(global_dist_history[i][point_ind, :].tocoo().col)

        # Plot
        for i in range(k):

            cur_x = i % num_subplts_per_row
            cur_y = i // num_subplts_per_row

            ax = fig.add_subplot(num_rows, num_cols, i+1, projection='3d')

            # Get neighbors and other points, diff-vec-neighbor etc.
            if plot_local_dists:
                dists = np.squeeze(sps.csr_matrix((np.linalg.norm(diff_vec_history[i], axis=1), (row, col)), shape=(self.n_pts, self.n_pts))[point_ind, neighbors_inds].toarray())
            else:
                neighbors_inds = neighbors_inds_list[i]
                dists = np.squeeze(global_dist_history[i][point_ind, neighbors_inds].toarray())

            mask = np.ones(self.n_pts, dtype=bool)
            mask[neighbors_inds] = False
            non_neighbor_inds = np.arange(self.n_pts)[mask]
            sizes = 70*np.ones(len(neighbors_inds))

            diff_vec_arr_point = diff_vec_history[i][pos[point_ind]:pos[point_ind+1], :]

            # Plot
            ax.scatter(self.data[non_neighbor_inds, 0], self.data[non_neighbor_inds, 1], self.data[non_neighbor_inds, 2], c='black', alpha=0.1, depthshade=True)
            if all((plot_new_positions, plot_local_dists)):
                ax.scatter(self.data[point_ind, 0] + diff_vec_arr_point[:, 0],
                           self.data[point_ind, 1] + diff_vec_arr_point[:, 1],
                           self.data[point_ind, 2] + diff_vec_arr_point[:, 2],
                           c=dists, s=sizes, cmap=plt.cm.hot, depthshade=False)
            else:
                ax.scatter(self.data[neighbors_inds, 0], self.data[neighbors_inds, 1], self.data[neighbors_inds, 2], c=dists, s=sizes, cmap=plt.cm.hot, depthshade=False)
            ax.scatter(self.data[point_ind, 0], self.data[point_ind, 1], self.data[point_ind, 2], c='blue', s=300, depthshade=True)
            if i > 0:
                # Gradient arrow
                grad_arr = self.gradients(self.embedding[:, i], 1.0, 0.05, normalized=True)
                arrow = self.Arrow3D([self.data[point_ind, 0], self.data[point_ind, 0] + grad_scale*grad_arr[point_ind, 0]],
                                [self.data[point_ind, 1], self.data[point_ind, 1] + grad_scale*grad_arr[point_ind, 1]],
                                [self.data[point_ind, 2], self.data[point_ind, 2] + grad_scale*grad_arr[point_ind, 2]],
                                mutation_scale=20, lw=3, arrowstyle="-|>", color="b")
                ax.add_artist(arrow)

            ax.view_init(view_a, view_b)
            ax.set_aspect('equal')
            ax.set_title(subtitles[i])

        plt.tight_layout()

        self.save_if_path_given(save_path)
        plt.show()
